import dnslib
import sys

#A resolver wrapper around dnslib.py
# stolen wholesale from https://github.com/TheRook/subbrute
# thanks Rook
class resolver:
    #Google's DNS servers are only used if zero resolvers are specified by the user.
    pos = 0
    rcode = ""
    wildcards = {}
    failed_code = False
    last_resolver = ""

    def __init__(self, nameservers = ['8.8.8.8','8.8.4.4']):
        self.nameservers = nameservers

    def query(self, hostname, query_type = 'ANY', name_server = False, use_tcp = True):
        ret = []
        response = None
        if name_server == False:
            name_server = self.get_ns()
        else:
            self.wildcards = {}
            self.failed_code = None
        self.last_resolver = name_server
        query = dnslib.DNSRecord.question(hostname, query_type.upper().strip())
        try:
            response_q = query.send(name_server, 53, use_tcp)
            if response_q:
                response = dnslib.DNSRecord.parse(response_q)
            else:
                raise IOError("Empty Response")
        except Exception as e:
            #IOErrors are all conditions that require a retry.
            raise IOError(str(e))
        if response:
            self.rcode = dnslib.RCODE[response.header.rcode]
            for r in response.rr:
                try:
                    rtype = str(dnslib.QTYPE[r.rtype])
                except:#Server sent an unknown type:
                    rtype = str(r.rtype)
                #Fully qualified domains may cause problems for other tools that use subbrute's output.
                rhost = str(r.rname).rstrip(".")
                ret.append((rhost, rtype, str(r.rdata)))
            #What kind of response did we get?
            if self.rcode not in ['NOERROR', 'NXDOMAIN', 'SERVFAIL', 'REFUSED']:
                trace('!Odd error code:', self.rcode, hostname, query_type)
            #Is this a perm error?  We will have to retry to find out.
            if self.rcode in ['SERVFAIL', 'REFUSED', 'FORMERR', 'NOTIMP', 'NOTAUTH']:
                raise IOError('DNS Failure: ' + hostname + " - " + self.rcode)
            #Did we get an empty body and a non-error code?
            elif not len(ret) and self.rcode != "NXDOMAIN":
                raise IOError("DNS Error - " + self.rcode + " - for:" + hostname)
        return ret

    def was_successful(self):
        ret = False
        if self.failed_code and self.rcode != self.failed_code:
            ret = True
        elif self.rcode == 'NOERROR':
            ret = True
        return ret

    def get_returncode(self):
        return self.rcode

    def get_ns(self):
        if self.pos >= len(self.nameservers):
            self.pos = 0
        ret = self.nameservers[self.pos]
        # we may have metadata on how this resolver fails
        try:
            ret, self.wildcards, self.failed_code = ret
        except:
            self.wildcards = {}
            self.failed_code = None
        self.pos += 1
        return ret

    def add_ns(self, resolver):
        if resolver:
            self.nameservers.append(resolver)

    def get_authoritative(self, hostname):
        ret = []
        while not ret and hostname.count(".") >= 1:
            try:
                trace("Looking for nameservers:", hostname)
                nameservers = self.query(hostname, 'NS')
            except IOError:#lookup failed.
                nameservers = []
            for n in nameservers:
                #A DNS server could return anything.
                rhost, record_type, record = n
                if record_type == "NS":
                    #Return all A records for this NS lookup.
                    a_lookup = self.query(record.rstrip("."), 'A')
                    for a_host, a_type, a_record in a_lookup:
                        ret.append(a_record)
                #If a nameserver wasn't found try the parent of this sub.
            hostname = hostname[hostname.find(".") + 1:]
        return ret

    def get_last_resolver(self):
        return self.last_resolver

#Toggle debug output
verbose = False
def trace(*args, **kwargs):
    if verbose:
        for a in args:
            sys.stderr.write(str(a))
            sys.stderr.write(" ")
        sys.stderr.write("\n")



import re
import logging
import Resolver
import tldextract


class DmarcRecord(object):

    def __init__(self, domain):
        self.domain = domain
        self.version = None
        self.policy = None
        self.pct = None
        self.rua = None
        self.ruf = None
        self.subdomain_policy = None
        self.dkim_alignment = None
        self.spf_alignment = None
        self.record = None

    def __str__(self):
        return self.record

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def _store_tag_data(self, tag_name, tag_value):
        if tag_name == "v":
            self.version = tag_value
        elif tag_name == "p":
            self.policy = tag_value
        elif tag_name == "pct":
            self.pct = tag_value
        elif tag_name == "rua":
            self.rua = tag_value
        elif tag_name == "ruf":
            self.ruf = tag_value
        elif tag_name == "sp":
            self.subdomain_policy = tag_value
        elif tag_name == "adkim":
            self.dkim_alignment = tag_value
        elif tag_name == "aspf":
            self.spf_alignment = tag_value

    def process_tags(self, dmarc_string):
        TAG_NAME, TAG_VALUE = (0, 1)
        tags = _extract_tags(dmarc_string)
        for tag in tags:
            self._store_tag_data(tag[TAG_NAME], tag[TAG_VALUE])

    def is_record_strong(self):
        record_strong = False
        if self.policy is not None and (self.policy == "reject" or self.policy == "quarantine"):
            record_strong = True

        if not record_strong:
            try:
                record_strong = self.is_org_domain_strong()
            except OrgDomainException:
                record_strong = False

        return record_strong

    def is_subdomain_policy_strong(self):
        if self.subdomain_policy is not None:
            return self.subdomain_policy == "reject" or self.subdomain_policy == "quarantine"

    def is_org_domain_strong(self):
        org_record = self.get_org_record()
        subdomain_policy_strong = org_record.is_subdomain_policy_strong()
        if subdomain_policy_strong is not None:
            return subdomain_policy_strong
        else:
            return org_record.is_record_strong()

    def get_org_record(self):
        org_domain = self.get_org_domain()
        if org_domain == self.domain:
            raise OrgDomainException
        else:
            return DmarcRecord.from_domain(org_domain)

    def get_org_domain(self):
        try:
            domain_parts = tldextract.extract(self.domain)
            return "%(domain)s.%(tld)s" % {'domain': domain_parts.domain, 'tld': domain_parts.suffix}
        except TypeError:
            return None

    @staticmethod
    def from_dmarc_string(dmarc_string, domain):
        if dmarc_string is not None:
            dmarc_record = DmarcRecord(domain)
            dmarc_record.record = dmarc_string
            dmarc_record.process_tags(dmarc_string)
            return dmarc_record
        else:
            return DmarcRecord(domain)

    @staticmethod
    def from_domain(domain):
        dmarc_string = get_dmarc_string_for_domain(domain)
        if dmarc_string is not None:
            return DmarcRecord.from_dmarc_string(dmarc_string, domain)
        else:
            return DmarcRecord(domain)


def _extract_tags(dmarc_record):
    dmarc_pattern = "(\w+)=(.*?)(?:; ?|$)"
    return re.findall(dmarc_pattern, dmarc_record)


def _match_dmarc_record(txt_record):
    dmarc_pattern = re.compile('^"?(v=DMARC[^"]*)"?')
    potential_dmarc_match = dmarc_pattern.match(str(txt_record))
    return potential_dmarc_match


def _find_record_from_answers(txt_records):
    dmarc_record = None
    for record in txt_records:
        potential_match = _match_dmarc_record(record[2])
        if potential_match is not None:
            dmarc_record = potential_match.group(1)
    return dmarc_record


def get_dmarc_string_for_domain(domain):
    try:
        txt_records = Resolver.resolver().query("_dmarc." + domain, query_type="TXT")
        return _find_record_from_answers(txt_records)
    except IOError:
        # This is returned usually as a NXDOMAIN, which is expected.
        return None
    except TypeError as error:
        logging.exception(error)
        return None


class OrgDomainException(Exception):
    pass



import re
import logging
import Resolver


class SpfRecord(object):

    def __init__(self, domain):
        self.version = None
        self.record = None
        self.mechanisms = None
        self.all_string = None
        self.domain = domain
        self.recursion_depth = 0

    def __str__(self):
        return self.record

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def get_redirected_record(self):
        if self.recursion_depth >= 10:
            return SpfRecord(self.get_redirect_domain())
        else:
            redirect_domain = self.get_redirect_domain()
            if redirect_domain is not None:
                redirect_record = SpfRecord.from_domain(redirect_domain)
                redirect_record.recursion_depth = self.recursion_depth + 1
                return redirect_record

    def get_redirect_domain(self):
        redirect_domain = None
        for mechanism in self.mechanisms:
            redirect_mechanism = re.match('redirect=(.*)', mechanism)
            if redirect_mechanism is not None:
                redirect_domain = redirect_mechanism.group(1)
        return redirect_domain

    def get_include_domains(self):
        include_domains = []
        for mechanism in self.mechanisms:
            include_mechanism = re.match('include:(.*)', mechanism)
            if include_mechanism is not None:
                include_domains.append(include_mechanism.group(1))
        return include_domains

    def get_include_records(self):
        if self.recursion_depth >= 10:
            return {}
        else:
            include_domains = self.get_include_domains()
            include_records = {}
            for domain in include_domains:
                try:
                    include_records[domain] = SpfRecord.from_domain(domain)
                    include_records[domain].recursion_depth = self.recursion_depth + 1
                except IOError as e:
                    logging.exception(e)
                    include_records[domain] = None
            return include_records

    def _is_all_mechanism_strong(self):
        strong_spf_all_string = True
        if self.all_string is not None:
            if not (self.all_string == "~all" or self.all_string == "-all"):
                strong_spf_all_string = False
        else:
            strong_spf_all_string = False
        return strong_spf_all_string

    def _is_redirect_mechanism_strong(self):
        redirect_domain = self.get_redirect_domain()

        if redirect_domain is not None:
            redirect_mechanism = SpfRecord.from_domain(redirect_domain)

            if redirect_mechanism is not None:
                return redirect_mechanism.is_record_strong()
            else:
                return False
        else:
            return False

    def _are_include_mechanisms_strong(self):
        include_records = self.get_include_records()
        for record in include_records:
            if include_records[record] is not None and include_records[record].is_record_strong():
                return True
        return False

    def is_record_strong(self):
        strong_spf_record = self._is_all_mechanism_strong()
        if strong_spf_record is False:

            redirect_strength = self._is_redirect_mechanism_strong()
            include_strength = self._are_include_mechanisms_strong()

            strong_spf_record = False

            if redirect_strength is True:
                strong_spf_record = True

            if include_strength is True:
                strong_spf_record = True
        return strong_spf_record

    @staticmethod
    def from_spf_string(spf_string, domain):
        if spf_string is not None:
            spf_record = SpfRecord(domain)
            spf_record.record = spf_string
            spf_record.mechanisms = _extract_mechanisms(spf_string)
            spf_record.version = _extract_version(spf_string)
            spf_record.all_string = _extract_all_mechanism(spf_record.mechanisms)
            return spf_record
        else:
            return SpfRecord(domain)

    @staticmethod
    def from_domain(domain):
        spf_string = get_spf_string_for_domain(domain)
        if spf_string is not None:
            return SpfRecord.from_spf_string(spf_string, domain)
        else:
            return SpfRecord(domain)


def _extract_version(spf_string):
    version_pattern = "^v=(spf.)"
    version_match = re.match(version_pattern, spf_string)
    if version_match is not None:
        return version_match.group(1)
    else:
        return None


def _extract_all_mechanism(mechanisms):
    all_mechanism = None
    for mechanism in mechanisms:
        if re.match(".all", mechanism):
            all_mechanism = mechanism
    return all_mechanism


def _find_unique_mechanisms(initial_mechanisms, redirected_mechanisms):
    return [x for x in redirected_mechanisms if x not in initial_mechanisms]


def _extract_mechanisms(spf_string):
    spf_mechanism_pattern = ("(?:((?:\+|-|~)?(?:a|mx|ptr|include"
                             "|ip4|ip6|exists|redirect|exp|all)"
                             "(?:(?::|=|/)?(?:\S*))?) ?)")
    spf_mechanisms = re.findall(spf_mechanism_pattern, spf_string)

    return spf_mechanisms


def _match_spf_record(txt_record):
    spf_pattern = re.compile('^"?(v=spf[^"]*)"?')
    potential_spf_match = spf_pattern.match(str(txt_record))
    return potential_spf_match


def _find_record_from_answers(txt_records):
    spf_record = None
    for record in txt_records:
        potential_match = _match_spf_record(record[2])
        if potential_match is not None:
            spf_record = potential_match.group(1)
    return spf_record


def get_spf_string_for_domain(domain):
    try:
        txt_records = Resolver.resolver().query(domain, query_type="TXT")
        return _find_record_from_answers(txt_records)
    except IOError as e:
        # This is returned usually as a NXDOMAIN, which is expected.
        return None
